unit AMDI;

interface

uses Classes,Forms;

const AMDICloseState:integer=-1;
      AMDIWaitState:integer=0;
type

  SFormInfo=record
    Form:TForm;
    State:integer;
    OldProc:TCloseEvent;
  end;

  TAMDIEngine=class(TComponent)
  private
    FThread:TThread;
    FOnExecute:TNotifyEvent;
    FFormInfo:array[1..100] of SFormInfo;
    FFormCount:integer;
    FExecuting:boolean;
    FTerminated:boolean;
    FPID:integer;
    FFC:TFormClass;
    function GetActiveForm:TForm;
    function GetActiveState:integer;
    procedure SetActiveState(S:integer);
    function GetState(W:TForm):integer;
    procedure SetState(W:TForm;S:integer);
    procedure SetPID(PID:integer);
    procedure DoCloseForm;
    procedure DoNewForm;
    procedure DoShowForm;
    procedure DefaultClose(Sender: TObject; var Action: TCloseAction);
    function FindFormInfo(W:TForm):integer;
  public
    property ActiveState:integer read GetActiveState write SetActiveState;
    property ActiveForm:TForm read GetActiveForm;
    property Terminated:boolean read FTerminated;
    property Executing:boolean read FExecuting;
    property States[AForm:TForm]:integer read GetState write SetState;
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure NewForm(AFormClass:TFormClass);
    procedure CloseForm;
    procedure ShowForm;
    function ExistsForm(AForm:TForm):boolean;
    procedure Execute;
    procedure Terminate;
  published
    property EngineID:integer read FPID write SetPID;
    property OnExecute:TNotifyEvent read FOnExecute write FOnExecute;
  end;

  TAMDIEnginesList=class(TComponent)
  private
    FList:TList;
    function GetState(W:TForm):integer;
    procedure SetState(W:TForm;S:integer);
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure TerminateAll;
    function FindByForm(AForm:TForm):TAMDIEngine;
    function FindByID(ID:integer):TAMDIEngine;
    property States[AForm:TForm]:integer read GetState write SetState;
  end;  

var AMDIEngines:TAMDIEnginesList;

implementation

uses Windows;

type TAMDIThread=class(TThread)
  private
    FOwner:TAMDIEngine;
  protected
    procedure Execute; override;
  public
    constructor Create(O:TAMDIEngine);
    destructor Destroy; override;
    procedure DoMethod(M:TThreadMethod);
  end;

var MaxPID:integer;

procedure TAMDIThread.Execute;
begin
  try
    try
      FOwner.FExecuting:=true;
      if assigned(FOwner.OnExecute) then
        FOwner.OnExecute(FOwner);
    finally
      FOwner.FThread:=nil;
      FOwner.FExecuting:=false;
      AMDIEngines.FList.Remove(FOwner);
    end
  except
    Application.HandleException(Application);
  end;
end;

constructor TAMDIThread.Create(O:TAMDIEngine);
begin
  inherited Create(True);
  FreeOnTerminate:=true;
  FOwner:=O;
end;
destructor TAMDIThread.Destroy;
begin
  inherited Destroy;
end;
procedure TAMDIThread.DoMethod(M:TThreadMethod);
begin
  Synchronize(M);
end;

procedure TAMDIEngine.SetPID(PID:integer);
begin
  FPID:=PID;
  if PID>MaxPID then
    MaxPID:=PID;
end;
function TAMDIEngine.GetActiveForm:TForm;
begin
  Result:=nil;
  if FFormCount>0 then
    Result:=FFormInfo[FFormCount].Form;
end;
function TAMDIEngine.GetActiveState:integer;
begin
  Result:=0;
  if FFormCount>0 then
    Result:=FFormInfo[FFormCount].State;
end;
procedure TAMDIEngine.SetActiveState(S:integer);
begin
  if FFormCount>0 then
    FFormInfo[FFormCount].State:=S;
end;
function TAMDIEngine.FindFormInfo(W:TForm):integer;
var i:integer;
begin
  Result:=0;
  for i:=FFormCount downto 1 do
    if FFormInfo[i].Form=W then
    begin
      Result:=i;
      break;
    end;
end;
function TAMDIEngine.GetState(W:TForm):integer;
var i:integer;
begin
  Result:=0;
  i:=FindFormInfo(W);
  if i>0 then
    Result:=FFormInfo[i].State;
end;
procedure TAMDIEngine.SetState(W:TForm;S:integer);
var i:integer;
begin
  i:=FindFormInfo(W);
  if i>0 then
    FFormInfo[i].State:=S;
end;
constructor TAMDIEngine.Create(AOwner:TComponent);
begin
  inherited Create(AOWner);
  FThread:=nil;
  FOnExecute:=nil;
  FFormCount:=0;
  FExecuting:=false;
  FTerminated:=false;
  inc(MaxPID);
  FPID:=MaxPID;
end;
destructor TAMDIEngine.Destroy;
begin
  if Executing then
  begin
    Terminate;
    if assigned(FThread) then
      FThread.WaitFor;
  end;
  inherited Destroy;
end;
procedure TAMDIEngine.DoCloseForm;
var i:integer;
    x:boolean;
begin
  if FFormCount>0 then
  begin
    x:=false;
    for i:=0 to Screen.FormCount-1 do
      if Screen.Forms[i]=FFormInfo[FFormCount].Form then
      begin
        x:=true;
      end;
    if x then
      FFormInfo[FFormCount].Form.Close;
    Application.ProcessMessages;
    dec(FFormCount);
  end;
  if FFormCount>0 then
  begin
    with FFormInfo[FFormCount].Form do
    begin
      Enabled:=true;
      if FormStyle<>fsMDIChild then
        FormStyle:=fsMDIChild;
      if WindowState=wsMinimized then
        WindowState:=wsNormal;
      Show;
      Application.ProcessMessages;
    end;
  end;
end;
procedure TAMDIEngine.DoShowForm;
begin
  if FFormCount>0 then
  begin
    with FFormInfo[FFormCount].Form do
    begin
      if FormStyle<>fsMDIChild then
        FormStyle:=fsMDIChild;
      if WindowState=wsMinimized then
        WindowState:=wsNormal;
      Show;
    end;
    Application.ProcessMessages;
  end;
end;
procedure TAMDIEngine.DoNewForm;
begin
  if FFormCount>0 then
    FFormInfo[FFormCount].Form.Enabled:=False;
  inc(FFormCount);
  FFormInfo[FFormCount].State:=0;
  Application.CreateForm(FFC,FFormInfo[FFormCount].Form);
  Application.ProcessMessages;
  with FFormInfo[FFormCount].Form do
  begin
    FFormInfo[FFormCount].OldProc:=OnClose;
    OnClose:=DefaultClose;
  end;
end;
procedure TAMDIEngine.DefaultClose(Sender: TObject; var Action: TCloseAction);
var OldProc:TCloseEvent;
    i:integer;
begin
  i:=FindFormInfo(TForm(Sender));
  if i>0 then
  begin
    oldProc:=FFormInfo[i].OldProc;
    Action:=caFree;
    if Assigned(OldProc) then
      OldProc(Sender,Action);
    if Action=caFree then
      FFormInfo[i].State:=AMDICloseState;
  end;
end;
procedure TAMDIEngine.NewForm(AFormClass:TFormClass);
begin
  FFC:=AFormClass;
  if (GetCurrentThreadID<>MainThreadID) and
      assigned(FThread) then
    TAMDIThread(FThread).DoMethod(DoNewForm)
  else
    DoNewForm;
end;
procedure TAMDIEngine.CloseForm;
begin
  if (GetCurrentThreadID<>MainThreadID) and
      assigned(FThread) then
    TAMDIThread(FThread).DoMethod(DoCloseForm)
  else
    DoCloseForm;
end;
procedure TAMDIEngine.ShowForm;
begin
  if (GetCurrentThreadID<>MainThreadID) and
      assigned(FThread) then
    TAMDIThread(FThread).DoMethod(DoShowForm)
  else
    DoShowForm;
end;
function TAMDIEngine.ExistsForm(AForm:TForm):boolean;
var i:integer;
begin
  Result:=false;
  for i:=FFormCount downto 1 do
    if FFormInfo[i].Form=AForm then
    begin
      Result:=true;
      break;
    end;
end;
procedure TAMDIEngine.Execute;
var W:TForm;
begin
  if FExecuting then
  begin
    w:=ActiveForm;
    if assigned(w) then
    begin
      if w.WindowState=wsMinimized then
        w.WindowState:=wsNormal;
      w.Show;
    end;
    exit;
  end;
  FTerminated:=false;
  FFormCount:=0;
  AMDIEngines.FList.Add(Self);
  FThread:=TAMDIThread.Create(Self);
  FThread.Resume;
end;
procedure TAMDIEngine.Terminate;
begin
  FTerminated:=true;
  if assigned(FThread) then
    FThread.Terminate;
end;

function TAMDIEnginesList.GetState(W:TForm):integer;
var P:TAMDIEngine;
begin
  Result:=0;
  P:=FindByForm(W);
  if assigned(P) then
    Result:=P.States[w];
end;
procedure TAMDIEnginesList.SetState(W:TForm;S:integer);
var P:TAMDIEngine;
begin
  P:=FindByForm(W);
  if assigned(P) then
    P.States[w]:=S;
end;
constructor TAMDIEnginesList.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FList:=TList.Create;
end;
destructor TAMDIEnginesList.Destroy;
begin
  FList.Destroy;
  inherited Destroy;
end;
procedure TAMDIEnginesList.TerminateAll;
var i:integer;
begin
  for i:=0 to FList.Count-1 do
    TAMDIEngine(FList.Items[i]).Terminate;
  while FList.Count>0 do
    Application.HandleMessage;
end;
function TAMDIEnginesList.FindByForm(AForm:TForm):TAMDIEngine;
var i:integer;
begin
  Result:=nil;
  for i:=0 to FList.Count-1 do
    if TAMDIEngine(FList.Items[i]).ExistsForm(AForm) then
    begin
      Result:=TAMDIEngine(FList.Items[i]);
      break;
    end;
end;
function TAMDIEnginesList.FindByID(ID:integer):TAMDIEngine;
var i:integer;
begin
  Result:=nil;
  for i:=0 to FList.Count-1 do
    if TAMDIEngine(FList.Items[i]).FPID=ID then
    begin
      Result:=TAMDIEngine(FList.Items[i]);
      break;
    end;
end;

initialization
  AMDIEngines:=TAMDIEnginesList.Create(nil);
  MaxPID:=0;
finalization
  AMDIEngines.Destroy;
end.
